from ._titanic import Titanic
from ._uci_datasets import (
    Adult,
    AirQuality,
    BankMarketing,
    HeartDisease,
    Iris,
    OnlineRetail,
    REV,
    WDBC,
    WineQuality,
)

__all__ = [
    "Adult",
    "AirQuality",
    "BankMarketing",
    "HeartDisease",
    "Iris",
    "LazyDataLoader",
    "OnlineRetail",
    "REV",
    "TabularDataLoader",
    "Titanic",
    "WDBC",
    "WineQuality",
]
